;;; Mudsync --- Live hackable MUD
;;; Copyright © 2016 Christopher Allan Webber <cwebber@dustycloud.org>
;;;
;;; This file is part of Mudsync.
;;;
;;; Mudsync is free software; you can redistribute it and/or modify it
;;; under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; Mudsync is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Mudsync.  If not, see <http://www.gnu.org/licenses/>.

(use-modules (8sync systems actors)
             (8sync systems actors debug)
             (8sync agenda)
             (ice-9 format)
             (ice-9 match)
             (gdbm)
             (oop goops))



;;; Networking
;;; ==========

(define %default-server #f)
(define %default-port 8889)

(define-class <network-manager> (<actor>)
  (server-socket #:accessor nm-server-socket)
  ;; mapping of client -> client-id
  (clients #:accessor nm-clients
           #:init-thunk make-hash-table)
  ;; send input to this actor
  (send-input-to #:getter nm-send-input-to
                 #:init-keyword #:send-input-to)
  (message-handler
   #:init-value
   (make-action-dispatch
    ((start-listening actor message)
     (nm-install-socket actor (message-ref message 'server %default-server)
                        (message-ref message 'port %default-port)))
    ((send-to-client actor message client data)
     (nm-send-to-client-id actor client data)))))

(define-method (nm-close-everything (nm <network-manager>) remove-from-agenda)
  "Shut it down!"
  ;; close all clients
  (hash-for-each
   (lambda (_ client)
     (close client)
     (if remove-from-agenda
         (8sync-port-remove client)))
   (nm-clients nm))
  ;; reset the clients list
  (set! (nm-clients) (make-hash-table))
  ;; close the server
  (close (nm-server-socket nm))
  (if remove-from-agenda
      (8sync-port-remove (nm-server-socket nm))))

;; Maximum number of backlogged connections when we listen
(define %maximum-backlog-conns 128)     ; same as SOMAXCONN on Linux 2.X,
                                        ; says the intarwebs

(define (nm-install-socket nm server port)
  "Install socket on SERVER with PORT"
  (let ((s (socket PF_INET  ; ipv4
                   SOCK_STREAM  ; two-way connection-based byte stream
                   0))
        (addr (if server
                  (inet-pton AF_INET server)
                  INADDR_LOOPBACK)))
    ;; Totally mimed from the Guile manual.  Not sure if we need this, but:
    ;; http://www.unixguide.net/network/socketfaq/4.5.shtml
    (setsockopt s SOL_SOCKET SO_REUSEADDR 1) ; reuse port even if port is busy
    ;; Connecting to a non-specific address:
    ;;   (bind s AF_INET INADDR_ANY port)
    ;; Should this be an option?  Guess I don't know why we'd need it
    ;; @@: If we wanted to support listening on a particular hostname,
    ;;   could see 8sync's irc.scm...
    (bind s AF_INET addr port)
    ;; Listen to connections
    (listen s %maximum-backlog-conns)

    ;; Throw a system-error rather than block on an (accept)
    ;; that has nothing to do
    (fcntl s F_SETFL
           (logior O_NONBLOCK
                   (fcntl s F_GETFL)))

    ;; @@: This is used in Guile's http server under the commit:
    ;;       * module/web/server/http.scm (http-open): Ignore SIGPIPE. Keeps the
    ;;         server from dying in some circumstances.
    ;;   (sigaction SIGPIPE SIG_IGN)
    ;; Will this break other things that use pipes for us though?

    (set! (nm-server-socket nm) s)

    (format #t "Listening for clients in pid: ~s\n" (getpid))
    (8sync-port s #:read (lambda (s) (nm-new-client nm s)))
    ;; TODO: set up periodic close of idle connections?
    ))

(define (nm-new-client nm s)
  "Handle new client coming in to socket S"
  (let* ((client-connection (accept s))
         (client-details (cdr client-connection))
         (client (car client-connection)))
    (format #t "New client: ~s\n" client-details)
    (format #t "Client address: ~s\n"
            (gethostbyaddr
             (sockaddr:addr client-details)))

    (let ((client-id (big-random-number)))
      (hash-set! (nm-clients nm) client-id client)
      ;; @@: Do we need an 8sync-port-wait here?
      ;;   Is such a thing even possible? :\
      (8sync-port client #:read (nm-make-client-receive nm client-id))
      (<- nm (nm-send-input-to nm) 'new-client #:client client-id))))

(define (nm-make-client-receive nm client-id)
  "Make a method to receive client data"
  (let ((buffer '()))
    (define (reset-buffer)
      (set! buffer '()))
    (define (should-read-char client)
      (and (not (port-closed? client))
           (char-ready? client)
           (not (eof-object? (peek-char client)))))
    (define (receive-handler client)
      (while (should-read-char client)
        (set! buffer (cons (read-char client) buffer))
        (match buffer
          (;; @@: Do we need the "char?"
           (#\newline #\return (? char? line-chars) ...)
           (let ((ready-line (list->string (reverse line-chars))))
             ;; reset buffer
             (set! buffer '())
             ;; run it
             (nm-handle-line nm client client-id ready-line)))
          (_ #f)))
      ;; Shut things down on closed port or EOF object
      (cond
       ((port-closed? client)
        (nm-handle-port-closed nm client client-id))
       ((and (char-ready? client)
             (eof-object? (peek-char client)))
        (nm-handle-port-eof nm client client-id))))
    receive-handler))

(define (nm-handle-port-closed nm client client-id)
  "Handle a closed port"
  (format #t "DEBUG: handled closed port ~x\n" client-id)
  (8sync-port-remove client)
  (hash-remove! (nm-clients nm) client-id))

(define-method (nm-handle-port-eof nm client client-id)
  "Handle seeing an EOF on port"
  (format #t "DEBUG: handled eof-object on port ~x\n" client-id)
  (close client)
  (8sync-port-remove client)
  (hash-remove! (nm-clients nm) client-id))

(define-method (nm-handle-line nm client client-id line)
  "Handle an incoming line of input from a client"
  (<- nm (nm-send-input-to nm) 'client-input
      #:data line
      #:client client-id))

(define-method (nm-send-to-client-id nm client-id data)
  "Send DATA to TO-CLIENT id"
  (define client-obj (hash-ref (nm-clients nm) client-id))
  (if (not client-obj)
      (throw 'no-such-client
             "Asked to send data to client but that client doesn't exist"
             #:client-id client-id
             #:data data))
  (display data client-obj))

; (ez-run-hive hive (list (bootstrap-message hive (actor-id nm) 'start-listening)))


;; (define-method (nm-close-port (nm <network-manager>)))



;;; The game master!  Runs the world.
;;; =================================

;; @@: We could call this a "world builder" instead...
;;   I kinda like calling it a GM though.

(define-class <game-master> (<actor>)
  ;; Directory of "special" objects.
  (special-dir #:init-thunk make-hash-table
               #:getter gm-special-dir)

  ;; Room directory.  Room symbols to locations.
  (room-dir #:init-thunk make-hash-table
            #:getter gm-room-dir)

  ;; A mapping of client ids to in-game actors
  ;; and a reverse ;p
  (client-dir #:init-thunk make-hash-table
              #:getter gm-client-dir)
  (reverse-client-dir #:init-thunk make-hash-table
                      #:getter gm-reverse-client-dir)

  ;; Network manager
  (network-manager #:accessor gm-network-manager
                   #:init-value #f)

  ;; How we get a new connection acclimated to the system
  (new-conn-handler #:accessor gm-new-conn-handler
                    #:init-keyword #:new-conn-handler)

  (message-handler
   #:init-value
   (make-action-dispatch
    (init-world (wrap-apply gm-init-world))
    (client-input (wrap-apply gm-handle-client-input))
    (lookup-room (wrap-apply gm-lookup-room))
    (new-client (wrap-apply gm-new-client))
    (write-home (wrap-apply gm-write-home)))))


;;; .. begin world init stuff ..

(define (gm-init-world gm message)
  ;; Load database
  ;;  TODO

  ;; Init basic rooms / structure
  (gm-init-rooms gm (message-ref message 'room-spec))

  ;; Restore database-based actors
  ;;  TODO

  ;; Set up the network
  (gm-setup-network gm))

(define (gm-init-rooms gm rooms-spec)
  "Initialize the prebuilt rooms"
  ;; @@: Would it be nicer to just allow passing in
  ;;     #:exits to the room spec itself?
  (define (exit-from-spec exit-spec)
    "Take room exits syntax from the spec, turn it into exits"
    (match exit-spec
      ((name to-symbol desc)
       (make <exit>
         #:name name
         #:to-symbol to-symbol
         #:desc desc))))

  (define rooms
    (map
     (match-lambda
       ((room-symbol room-class
                     room-args ...
                     (room-exits ...))
        ;; initialize the room
        (let ((room
               (apply create-actor* gm room-class "room"
                      #:gm (actor-id gm)
                      #:exits (map exit-from-spec room-exits)
                      room-args)))
          ;; register the room
          (hash-set! (gm-room-dir gm) room-symbol room)
          ;; pass it back to the map
          room)))
     rooms-spec))

  ;; now wire up all the exits
  (for-each
   (lambda (room)
     (format #t "Wiring up ~s...\n" (address->string room))
     (<-wait gm room 'wire-exits!))
   rooms))


(define (gm-setup-network gm)
  ;; Create a default network manager if none available
  (set! (gm-network-manager gm)
        (create-actor* gm <network-manager> "netman"
                       #:send-input-to (actor-id gm)))

  ;; TODO: Add host and port options
  (<-wait gm (gm-network-manager gm) 'start-listening))

(define (gm-setup-database gm)
  'TODO)

;;; .. end world init stuff ...

(define-mhandler (gm-new-client actor message client)
  ;; @@: Maybe more indirection than needed for this
  ((gm-new-conn-handler actor) actor client))


(define (gm-handle-client-input actor message)
  "Handle input from a client."
  (define client-id (message-ref message 'client))
  (define input (message-ref message 'data))
  (format #t "From ~s: ~s\n" client-id input)
  (<- actor (gm-network-manager actor) 'send-to-client
      #:client client-id
      #:data "Thanks, we got it!\n"))

(define-mhandler (gm-lookup-room actor message symbol)
  (define room-id
    (slot-ref (gm-room-dir actor) symbol))
  (<-reply actor message room-id))

(define-mhandler (gm-write-home actor message text)
  (define client-id (hash-ref (gm-reverse-client-dir actor)
                              (message-from message)))
  (<- actor (gm-network-manager actor) 'send-to-client
      #:client client-id
      #:data text))


;;; GM utilities

(define (gm-register-client! gm client-id player)
  (hash-set! (gm-client-dir gm) client-id player)
  (hash-set! (gm-reverse-client-dir gm) player client-id))

(define (gm-unregister-client! gm client-id)
  "Remove a connection/player combo and ask them to self destruct"
  (match (hash-remove! (gm-client-dir gm) client-id)  ; Remove from our client dir
    ((_ . player-id)
     ;; Remove from reverse table too
     (hash-remove! (gm-reverse-client-dir gm) client-id)
     ;; Destroy player 
     (<- gm player-id 'destroy-self))
    (#f (throw 'no-client-to-unregister
               "Can't unregister a client that doesn't exist?"
               client-id))))

;;; An easy default

(define (make-default-room-conn-handler default-room)
  "Make a handler for a GM that dumps people in a default room
with an anonymous persona"
  (let ((count 0))
    (lambda (gm client-id)
      (define guest-name (string-append "Guest-"
                                        (number->string count)))
      (define room-id
        (hash-ref (gm-room-dir gm) default-room))
      ;; create and register the player
      (define player
        (create-actor* gm <player> "player"
                       #:username guest-name
                       #:gm (actor-id gm)
                       #:client client-id))

      ;; Register the player in our database of players -> connections
      (gm-register-client! gm client-id player)
      ;; Dump the player into the default room
      (<-wait gm player 'set-loc! #:id room-id)
      ;; Initialize the player
      (<- gm player 'init))))


;;; Game actor
;;; ==========

;;; *all* game components that talk to players should somehow
;;; derive from this class.
;;; And all of them need a GM!

(define-class <game-actor> (<actor>)
  (gm #:init-keyword #:gm
      #:getter game-actor-gm))



;;; Rooms
;;; =====

;; @@: Maybe make this into a record type when this congeals a bit?
;;   I dunno?

(define-class <exit> ()
  ;; Used for wiring
  (to-symbol #:accessor exit-to-symbol
             #:init-keyword #:to-symbol)
  ;; The actual address we use
  (to-address #:accessor exit-to-address
              #:init-keyword #:address)
  ;; Name of the room (@@: Should this be names?)
  (name #:accessor exit-name
        #:init-keyword #:name)
  (desc #:accessor exit-desc
               #:init-keyword #:desc)

  ;; *Note*: These two methods have an extra layer of indirection, but
  ;;   it's for a good reason.
  (visible-check #:init-value (const #t)
                 #:init-keyword #:visible-check)
  ;; By default all exits can be traversed
  (traverse-check #:init-value (const #t)
                  #:init-keyword #:traverse-check))

(define* (exit-can-traverse? exit actor
                             #:optional (target-actor (actor-id actor)))
  ((slot-ref exit 'traverse-check) exit actor target-actor))

(define* (exit-is-visible? exit actor
                           #:optional (target-actor (actor-id actor)))
  ((slot-ref exit 'traverse-check) exit actor target-actor))


;; Kind of a useful utility, maybe?
(define (simple-slot-getter slot)
  (lambda (actor message)
    (reply-message actor message
                   #:val (slot-ref actor slot))))


(define-class <room> (<game-actor>)
  (name #:init-keyword #:name)
  (desc #:init-value ""
               #:init-keyword #:desc)
  ;; Uses a hash table like a set (values ignored)
  (occupants #:init-thunk make-hash-table)
  ;; A list of <exit>
  (exits #:init-value '()
         #:getter room-exits)
  ;; @@: Maybe eventually <room> will inherit from some more general
  ;;  game object class

  (message-handler
   #:allocation #:each-subclass
   #:init-value
   (make-action-dispatch
    ;; desc == description
    (get-desc
     (simple-slot-getter 'desc))
    (get-name
     (simple-slot-getter 'name))
    ((register-occupant! actor message who)
     "Register an actor as being a occupant of this room"
     (hash-set! (slot-ref actor 'occupants) who #t))
    ((evict-occupant! actor message who)
     "De-register an occupant removed from the room"
     (hash-remove! (slot-ref actor 'occupants) who))
    (wire-exits! (wrap-apply room-wire-exits!)))))

(define (room-wire-exits! room message)
  "Actually hook up the rooms' exit addresses to the rooms they
claim to point to."
  (for-each
   (lambda (exit)
     (define new-exit
       (<-wait room (game-actor-gm room) 'lookup-room
               #:symbol (exit-to-symbol exit)))

     (set! (exit-to-address exit) new-exit))

   (room-exits room)))



;;; Players
;;; =======

(define-class <player> (<game-actor>)
  (username #:init-keyword #:username
            #:accessor player-username)
  ;; location id
  (loc #:init-value #f
       #:accessor player-loc)
  ;; Connection id
  (client #:accessor player-client)

  (message-handler
   #:init-value
   (make-action-dispatch
    (set-loc! (wrap-apply player-set-loc!))
    (init (wrap-apply player-init!)))))

;;; player message handlers

(define-mhandler (player-set-loc! player message id)
  (format #t "DEBUG: Location set to ~s for player ~s"
          id (actor-id-actor player))
  (set! (player-loc player) id))

(define-mhandler (player-init! player message)
  (player-look-around player))

;;; player methods

(define (player-look-around player)
  (define room-name
    (message-ref
     (<-wait player (player-loc player) 'get-name)
     'val))
  (define room-desc
    (message-ref
     (<-wait player (player-loc player) 'get-desc)
     'val))
  (define message-text
    (format #f "**~a**\n~a\n" room-name room-desc))

  (<- player (game-actor-gm player) 'write-home #:text message-text))


;;; Debugging stuff
;;; ===============

(define %test-gm #f)

(define (run-demo db-path room-spec default-room)
  (define hive (make-hive))
  (define new-conn-handler
    (make-default-room-conn-handler default-room))
  (define gm
    (hive-create-actor-gimmie* hive <game-master> "gm"
                               #:new-conn-handler new-conn-handler))
  (set! %test-gm gm)
  ;; @@: Boy, wouldn't it be nice if the agenda could do things
  ;;   on interrupt :P
  (ez-run-hive hive
               (list (bootstrap-message hive (actor-id gm) 'init-world
                                        #:room-spec room-spec))))
